# Generated by Django 4.1.7 on 2023-03-16 20:06

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [
        ("onto", "0004_remove_entity_created_time_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="entity",
            name="created_time",
            field=models.DateTimeField(
                auto_now_add=True,
                default=django.utils.timezone.now,
                help_text="When this entity and associated object were created.",
            ),
            preserve_default=False,
        ),
    ]
