# Generated by Django 4.1.7 on 2023-03-16 18:43

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("onto", "0001_initial"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="domain",
            name="entity",
        ),
        migrations.RemoveField(
            model_name="membership",
            name="domain",
        ),
        migrations.RemoveField(
            model_name="membership",
            name="entity",
        ),
        migrations.RemoveField(
            model_name="entity",
            name="domains",
        ),
        migrations.AlterField(
            model_name="entity",
            name="xattrs",
            field=models.JSONField(
                blank=True,
                default=dict,
                help_text="Arbitrary JSON attributes, e.g. tags, to attach to this object.",
                verbose_name="extended attributes",
            ),
        ),
    ]
