# Generated by Django 4.1.7 on 2023-03-16 18:43

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("onto_authz", "0002_domain_membership_domain_entities_and_more"),
        ("onto", "0002_remove_domain_entity_remove_membership_domain_and_more"),
    ]

    operations = [
        migrations.DeleteModel(
            name="Domain",
        ),
        migrations.DeleteModel(
            name="Membership",
        ),
    ]
