# Generated by Django 4.1.7 on 2023-03-16 20:05

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("onto", "0003_delete_domain_delete_membership"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="entity",
            name="created_time",
        ),
        migrations.RemoveField(
            model_name="entity",
            name="updated_time",
        ),
    ]
