# Generated by Django 4.1.7 on 2023-03-16 18:51

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("onto_authz", "0002_domain_membership_domain_entities_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="domain",
            name="active",
            field=models.BooleanField(
                default=True,
                help_text="When inactive, this domain cannot grant authorization.",
            ),
        ),
    ]
